/***********************************************************************/
/*                                                                     */
/*   svm_learn_main.c                                                  */
/*                                                                     */
/*   Command line interface to the learning module of the              */
/*   Support Vector Machine.                                           */
/*                                                                     */
/*   Author: Thorsten Joachims                                         */
/*   Date: 02.07.02                                                    */
/*                                                                     */
/*   Copyright (c) 2000  Thorsten Joachims - All rights reserved       */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/


/* if svm-learn is used out of C++, define it as extern "C" */
#ifdef __cplusplus
extern "C" {
#endif

# include "svm_common.h"
# include "svm_learn.h"

#ifdef __cplusplus
}
#endif

char docfile[200];           /* file with training examples */
char modelfile[200];         /* file for resulting classifier */
char restartfile[200];       /* file with initial alphas */

void   read_input_parameters(int, char **, char *, char *, char *, long *, 
                             LEARN_PARM *, KERNEL_PARM *);
void   wait_any_key();
void   print_help();

/* Razvan: begin */
void read_devel_docs(char *docfile, DOC ***docs, double **label, 
                     long int totwords, long int *totdoc);
void svm_free_data(MODEL* model, DOC** docs, double* target, long totdoc,
                   DOC** docs_devel, double* target_devel, long totd_devel);
void svm_one_class_train(int argc, char* argv[], MODEL** model,
                         LEARN_PARM* learn_parm, KERNEL_PARM* kernel_parm,
                         DOC ***docs, double **target,
                         long *totwords, long *totdoc,
                         DOC ***docs_devel, double **target_devel,
                         long *totd_devel, double factor, int ratio);
void svm_one_class_train_all(int argc, char* argv[], MODEL** model,
                             LEARN_PARM* learn_parm, KERNEL_PARM* kernel_parm,
                             DOC ***docs, double **target,
                             long *totwords, long *totdoc,
                             DOC ***docs_devel, double **target_devel,
                             long *totd_devel, double factor, int ratio);
extern int terminated;

extern double *primal, *dual;
extern long precision_violations;
extern double opt_precision;
extern long   maxiter;
extern double lindep_sensitivity;
extern long  smallroundcount;
extern long  roundnumber;

void refresh_optimization_vars()
{
  primal=0; dual=0;
  precision_violations=0;
  opt_precision=1E-5;
  maxiter=200;
  lindep_sensitivity=1E-8;
  smallroundcount=0;
  roundnumber=0;
}
/* Razvan: end */



int main (int argc, char* argv[])
{
  DOC **docs;  /* training examples */
  long totwords,totdoc,i;
  double *target;
  double *alpha_in=NULL;
  LEARN_PARM learn_parm;
  KERNEL_PARM kernel_parm;
  MODEL *model;
  
  /* Razvan: begin */
  DOC **docs_devel;
  long totd_devel;
  double *target_devel;

  long j, k, labeled, nterm;
  double dist;
  long ucount, lcount, total;

  long max_iter = 20;
  double fms[max_iter];
  double max_factor;
  double factor, svm_c;
  double max_fm;
  int ratio;
  double max_ratio;
  int RATIO = 10;
  /* Razvan: end */

  terminated = 0;
  max_fm = -1;

  for (ratio = 1; ratio <= RATIO; ratio++) {
    printf("Training with ratio = %d\n", ratio);
    k = 0;
    nterm = 0;
    factor = 1.0;
   
    do {
      printf("Iteration %ld\n", k);
      svm_one_class_train(argc, argv, &model,
                          &learn_parm, &kernel_parm,
                          &docs, &target,
                          &totwords, &totdoc,
                          &docs_devel, &target_devel,
                          &totd_devel, factor, ratio);
      svm_c = learn_parm.svm_c;
    
      if (terminated) {
        printf("TERMINATED!\n");
        fms[k] = -1;
        nterm++;
      }
      else {
        nterm = 0;
        if (k == 0)
          write_model("first_svm_model", model);

        add_weight_vector_to_linear_model(model);

        /* Compute f-measure estimate */
        labeled = 0;
        lcount = ucount = 0;
        for (j = 0; j < totd_devel; j++) {
          if (target_devel[j] > 0) {
            labeled += 1;
          }
          dist = classify_example_linear(model, docs_devel[j]);
          if (dist > 0) {
            if (target_devel[j] > 0)
              lcount++;
            else
              ucount++;
          }
        }
        fms[k] =
          (totd_devel / (double) (lcount + ucount)) *
          (lcount / (double) labeled) *
          (lcount / (double) labeled);
        /*
          fms[k] =
          lcount / (double) (labeled) - (lcount + ucount) / (double) n;
        */
        if (fms[k] > max_fm) {
          max_fm = fms[k];
          max_factor = factor;
          max_ratio = ratio;
        }
        k++;
      }
    
      svm_free_data(model, docs, target, totdoc,
                    docs_devel, target_devel, totd_devel);

      factor *= 2.0;
    }
    while ((svm_c < 256 || k < max_iter) && nterm < 5);

    /* Print f-measures for each C parameter */
    svm_c = 1.0 / totdoc;
    for (j = 0; j < k; j++) {
      printf("c = %.8g fm = %.8g.\n", svm_c, fms[j]);
      svm_c *= 2.0;
    }
  }
  
  if (max_fm > 0) {
    /* Retrain second SVM model with best C and J parameters */
    printf("Retraining for max devel fm =  %.8g\n", max_fm);
    svm_one_class_train(argc, argv, &model,
                        &learn_parm, &kernel_parm,
                        &docs, &target,
                        &totwords, &totdoc,
                        &docs_devel, &target_devel,
                        &totd_devel, max_factor, max_ratio);
    write_model("second_svm_model", model);
    svm_free_data(model, docs, target, totdoc,
                  docs_devel, target_devel, totd_devel);

    /* Train SVM model on all (train + devel)  data with best C and J parameters */
    printf("Retraining on train + devel for max devel fm =  %.8g\n", max_fm);
    svm_one_class_train_all(argc, argv, &model,
                            &learn_parm, &kernel_parm,
                            &docs, &target,
                            &totwords, &totdoc,
                            &docs_devel, &target_devel,
                            &totd_devel, max_factor, max_ratio);
    write_model("third_svm_model", model);
    svm_free_data(model, docs, target, totdoc,
                  docs_devel, target_devel, totd_devel);
  }
  
  return 0;
}


void svm_free_data(MODEL* model, DOC** docs, double* target, long totdoc,
                   DOC** docs_devel, double* target_devel, long totd_devel)
{
  int i;
  
  free_model(model,0);

  for(i=0;i<totdoc;i++) 
    free_example(docs[i],1);
  free(docs);
  free(target);

  for(i=0;i<totd_devel;i++) 
    free_example(docs_devel[i],1);
  free(docs_devel);
  free(target_devel);
}


void svm_one_class_train(int argc, char* argv[], MODEL** model,
                         LEARN_PARM* learn_parm, KERNEL_PARM* kernel_parm,
                         DOC ***docs, double **target,
                         long *totwords, long *totdoc,
                         DOC ***docs_devel, double **target_devel,
                         long *totd_devel, double factor, int ratio)
{
  long n, n1, n2, i;
  double* ntarget;

  /* Read training & development examples */
  refresh_optimization_vars();
  *model = (MODEL *) my_malloc(sizeof(MODEL));
  read_input_parameters(argc,argv,docfile,modelfile,restartfile,&verbosity,
                        learn_parm, kernel_parm);
  read_documents(docfile, docs, target, totwords, totdoc);
  read_devel_docs(kernel_parm->custom, docs_devel, target_devel,
                  *totwords, totd_devel);

  /* Set initial cost factors */
  n = *totdoc;
  ntarget = *target;
  n1 = n2 = 0;
  for (i = 0; i < n; i++) {
    if (ntarget[i] > 0)
      n1++;
    else
      n2++;
  }

  learn_parm->svm_costratio = ratio * n2 / (double) n1;
  learn_parm->svm_c = factor / n;
  printf("SVM parameters: c = %.8g ratio (j) = %.8g.\n",
         learn_parm->svm_c, learn_parm->svm_costratio);

  svm_learn_classification(*docs, *target, *totdoc, *totwords,
                           learn_parm, kernel_parm,
                           NULL, *model, NULL);
}


void svm_one_class_train_all(int argc, char* argv[], MODEL** model,
                             LEARN_PARM* learn_parm, KERNEL_PARM* kernel_parm,
                             DOC ***docs, double **target,
                             long *totwords, long *totdoc,
                             DOC ***docs_devel, double **target_devel,
                             long *totd_devel, double factor, int ratio)
{
  double* ntarget;
  DOC** docs_all;
  double* class_all;
  long totd_all;
  long n, n1, n2, i;

  /* Read training & development examples */
  refresh_optimization_vars();
  *model = (MODEL *) my_malloc(sizeof(MODEL));
  read_input_parameters(argc,argv,docfile,modelfile,restartfile,&verbosity,
                        learn_parm, kernel_parm);
  read_documents(docfile, docs, target, totwords, totdoc);
  read_devel_docs(kernel_parm->custom, docs_devel, target_devel,
                  *totwords, totd_devel);

  /* Set initial cost factors */
  ntarget = *target;
  n = *totdoc;
  n1 = n2 = 0;
  for (i = 0; i < n; i++) {
    if (ntarget[i] > 0)
      n1++;
    else
      n2++;
  }

  totd_all = *totdoc + *totd_devel;
  docs_all = (DOC**) my_malloc(totd_all * sizeof(DOC*));
  class_all = (double*) my_malloc(totd_all * sizeof(double));
  for (i = 0; i < n; i++) {
    docs_all[i] = (*docs)[i];
    class_all[i] = (*target)[i];
  }
  for (i = 0; i < *totd_devel; i++) {
    docs_all[n + i] = (*docs_devel)[i];
    class_all[n + i] = (*target_devel)[i];
  }
  
  learn_parm->svm_costratio = ratio * n2 / (double) n1;
  learn_parm->svm_c = factor / totd_all;
  printf("SVM parameters: c = %.8g ratio (j) = %.8g.\n",
         learn_parm->svm_c, learn_parm->svm_costratio);

  svm_learn_classification(docs_all, class_all, totd_all, *totwords,
                           learn_parm, kernel_parm,
                           NULL, *model, NULL);

  free(docs_all);
  free(class_all);
}


/* Razvan: begin */
void read_devel_docs(char *docfile, DOC ***docs, double **label, 
                    long int totwords, long int *totdoc)
{
  char *line,*comment;
  WORD *words;
  long dnum=0,wpos,dpos=0,dneg=0,dunlab=0,queryid,slackid,max_docs;
  long max_words_doc, ll;
  double doc_label,costfactor;
  FILE *docfl;
  int j;

  nol_ll(docfile,&max_docs,&max_words_doc,&ll); /* scan size of input file */
  max_words_doc+=2;
  ll+=2;
  max_docs+=2;

  (*docs) = (DOC **)my_malloc(sizeof(DOC *)*max_docs);    /* feature vectors */
  (*label) = (double *)my_malloc(sizeof(double)*max_docs); /* target values */
  line = (char *)my_malloc(sizeof(char)*ll);

  if ((docfl = fopen (docfile, "r")) == NULL)
    { perror (docfile); exit (1); }

  words = (WORD *)my_malloc(sizeof(WORD)*(max_words_doc+10));
  dnum=0;
  while((!feof(docfl)) && fgets(line,(int)ll,docfl)) {
    if(line[0] == '#') continue;  /* line contains comments */
    parse_document(line,words,&doc_label,&queryid,&slackid,&costfactor,
                   &wpos,max_words_doc,&comment);

    (*label)[dnum]=doc_label;
    /* printf("docnum=%ld: Class=%f ",dnum,doc_label); */
    for(j=0;(words[j]).wnum != 0;j++) {     /* check if feature numbers   */
      if((words[j]).wnum > totwords)        /* are not larger than in     */
        (words[j]).wnum=0;                  /* model. Remove feature if   */
    }                                       /* necessary.                 */

    (*docs)[dnum] = create_example(dnum,queryid,slackid,costfactor,
                                   create_svector(words,comment,1.0));
    /* printf("\nNorm=%f\n",((*docs)[dnum]->fvec)->twonorm_sq);  */
    dnum++; 
  } 

  fclose(docfl);
  free(line);
  free(words);
  (*totdoc)=dnum;
}
/* Razvan: end */


/*---------------------------------------------------------------------------*/

void read_input_parameters(int argc,char *argv[],char *docfile,char *modelfile,
			   char *restartfile,long *verbosity,
			   LEARN_PARM *learn_parm,KERNEL_PARM *kernel_parm)
{
  long i;
  char type[100];
  
  /* set default */
  set_learning_defaults(learn_parm, kernel_parm);
  strcpy (modelfile, "svm_model");
  strcpy (restartfile, "");
  (*verbosity)=1;
  strcpy(type,"c");

  for(i=1;(i<argc) && ((argv[i])[0] == '-');i++) {
    switch ((argv[i])[1]) 
      { 
      case '?': print_help(); exit(0);
      case 'z': i++; strcpy(type,argv[i]); break;
      case 'v': i++; (*verbosity)=atol(argv[i]); break;
      case 'b': i++; learn_parm->biased_hyperplane=atol(argv[i]); break;
      case 'i': i++; learn_parm->remove_inconsistent=atol(argv[i]); break;
      case 'f': i++; learn_parm->skip_final_opt_check=!atol(argv[i]); break;
      case 'q': i++; learn_parm->svm_maxqpsize=atol(argv[i]); break;
      case 'n': i++; learn_parm->svm_newvarsinqp=atol(argv[i]); break;
      case '#': i++; learn_parm->maxiter=atol(argv[i]); break;
      case 'h': i++; learn_parm->svm_iter_to_shrink=atol(argv[i]); break;
      case 'm': i++; learn_parm->kernel_cache_size=atol(argv[i]); break;
      case 'c': i++; learn_parm->svm_c=atof(argv[i]); break;
      case 'w': i++; learn_parm->eps=atof(argv[i]); break;
      case 'p': i++; learn_parm->transduction_posratio=atof(argv[i]); break;
      case 'j': i++; learn_parm->svm_costratio=atof(argv[i]); break;
      case 'e': i++; learn_parm->epsilon_crit=atof(argv[i]); break;
      case 'o': i++; learn_parm->rho=atof(argv[i]); break;
      case 'k': i++; learn_parm->xa_depth=atol(argv[i]); break;
      case 'x': i++; learn_parm->compute_loo=atol(argv[i]); break;
      case 't': i++; kernel_parm->kernel_type=atol(argv[i]); break;
      case 'd': i++; kernel_parm->poly_degree=atol(argv[i]); break;
      case 'g': i++; kernel_parm->rbf_gamma=atof(argv[i]); break;
      case 's': i++; kernel_parm->coef_lin=atof(argv[i]); break;
      case 'r': i++; kernel_parm->coef_const=atof(argv[i]); break;
      case 'u': i++; strcpy(kernel_parm->custom,argv[i]); break;
      case 'l': i++; strcpy(learn_parm->predfile,argv[i]); break;
      case 'a': i++; strcpy(learn_parm->alphafile,argv[i]); break;
      case 'y': i++; strcpy(restartfile,argv[i]); break;
      default: printf("\nUnrecognized option %s!\n\n",argv[i]);
	       print_help();
	       exit(0);
      }
  }
  if(i>=argc) {
    printf("\nNot enough input parameters!\n\n");
    wait_any_key();
    print_help();
    exit(0);
  }
  strcpy (docfile, argv[i]);
  if((i+1)<argc) {
    strcpy (modelfile, argv[i+1]);
  }
  if(learn_parm->svm_iter_to_shrink == -9999) {
    if(kernel_parm->kernel_type == LINEAR) 
      learn_parm->svm_iter_to_shrink=2;
    else
      learn_parm->svm_iter_to_shrink=100;
  }
  if(strcmp(type,"o")==0) {
    learn_parm->type=ONECLASS; /* Razvan */
  }
  else if(strcmp(type,"c")==0) {
    learn_parm->type=CLASSIFICATION;
  }
  else if(strcmp(type,"r")==0) {
    learn_parm->type=REGRESSION;
  }
  else if(strcmp(type,"p")==0) {
    learn_parm->type=RANKING;
  }
  else if(strcmp(type,"o")==0) {
    learn_parm->type=OPTIMIZATION;
  }
  else if(strcmp(type,"s")==0) {
    learn_parm->type=OPTIMIZATION;
    learn_parm->sharedslack=1;
  }
  else {
    printf("\nUnknown type '%s': Valid types are 'c' (classification), 'r' regession, and 'p' preference ranking.\n",type);
    wait_any_key();
    print_help();
    exit(0);
  }    
  if (!check_learning_parms(learn_parm, kernel_parm)) {
     wait_any_key();
     print_help();
     exit(0);
  }
}

void wait_any_key()
{
  printf("\n(more)\n");
  (void)getc(stdin);
}

void print_help()
{
  printf("\nSVM-light %s: Support Vector Machine, learning module     %s\n",VERSION,VERSION_DATE);
  copyright_notice();
  printf("   usage: svm_learn [options] example_file model_file\n\n");
  printf("Arguments:\n");
  printf("         example_file-> file with training data\n");
  printf("         model_file  -> file to store learned decision rule in\n");

  printf("General options:\n");
  printf("         -?          -> this help\n");
  printf("         -v [0..3]   -> verbosity level (default 1)\n");
  printf("Learning options:\n");
  printf("         -z {c,r,p}  -> select between classification (c), regression (r),\n");
  printf("                        and preference ranking (p) (default classification)\n");
  printf("         -c float    -> C: trade-off between training error\n");
  printf("                        and margin (default [avg. x*x]^-1)\n");
  printf("         -w [0..]    -> epsilon width of tube for regression\n");
  printf("                        (default 0.1)\n");
  printf("         -j float    -> Cost: cost-factor, by which training errors on\n");
  printf("                        positive examples outweight errors on negative\n");
  printf("                        examples (default 1) (see [4])\n");
  printf("         -b [0,1]    -> use biased hyperplane (i.e. x*w+b>0) instead\n");
  printf("                        of unbiased hyperplane (i.e. x*w>0) (default 1)\n");
  printf("         -i [0,1]    -> remove inconsistent training examples\n");
  printf("                        and retrain (default 0)\n");
  printf("Performance estimation options:\n");
  printf("         -x [0,1]    -> compute leave-one-out estimates (default 0)\n");
  printf("                        (see [5])\n");
  printf("         -o ]0..2]   -> value of rho for XiAlpha-estimator and for pruning\n");
  printf("                        leave-one-out computation (default 1.0) (see [2])\n");
  printf("         -k [0..100] -> search depth for extended XiAlpha-estimator \n");
  printf("                        (default 0)\n");
  printf("Transduction options (see [3]):\n");
  printf("         -p [0..1]   -> fraction of unlabeled examples to be classified\n");
  printf("                        into the positive class (default is the ratio of\n");
  printf("                        positive and negative examples in the training data)\n");
  printf("Kernel options:\n");
  printf("         -t int      -> type of kernel function:\n");
  printf("                        0: linear (default)\n");
  printf("                        1: polynomial (s a*b+c)^d\n");
  printf("                        2: radial basis function exp(-gamma ||a-b||^2)\n");
  printf("                        3: sigmoid tanh(s a*b + c)\n");
  printf("                        4: user defined kernel from kernel.h\n");
  printf("         -d int      -> parameter d in polynomial kernel\n");
  printf("         -g float    -> parameter gamma in rbf kernel\n");
  printf("         -s float    -> parameter s in sigmoid/poly kernel\n");
  printf("         -r float    -> parameter c in sigmoid/poly kernel\n");
  printf("         -u string   -> parameter of user defined kernel\n");
  printf("Optimization options (see [1]):\n");
  printf("         -q [2..]    -> maximum size of QP-subproblems (default 10)\n");
  printf("         -n [2..q]   -> number of new variables entering the working set\n");
  printf("                        in each iteration (default n = q). Set n < q to \n");
  printf("                        prevent zig-zagging.\n");
  printf("         -m [5..]    -> size of cache for kernel evaluations in MB (default 40)\n");
  printf("                        The larger the faster...\n");
  printf("         -e float    -> eps: Allow that error for termination criterion\n");
  printf("                        [y [w*x+b] - 1] >= eps (default 0.001)\n");
  printf("         -y [0,1]    -> restart the optimization from alpha values in file\n");
  printf("                        specified by -a option. (default 0)\n");
  printf("         -h [5..]    -> number of iterations a variable needs to be\n"); 
  printf("                        optimal before considered for shrinking (default 100)\n");
  printf("         -f [0,1]    -> do final optimality check for variables removed\n");
  printf("                        by shrinking. Although this test is usually \n");
  printf("                        positive, there is no guarantee that the optimum\n");
  printf("                        was found if the test is omitted. (default 1)\n");
  printf("         -y string   -> if option is given, reads alphas from file with given\n");
  printf("                        and uses them as starting point. (default 'disabled')\n");
  printf("         -# int      -> terminate optimization, if no progress after this\n");
  printf("                        number of iterations. (default 100000)\n");
  printf("Output options:\n");
  printf("         -l string   -> file to write predicted labels of unlabeled\n");
  printf("                        examples into after transductive learning\n");
  printf("         -a string   -> write all alphas to this file after learning\n");
  printf("                        (in the same order as in the training set)\n");
  wait_any_key();
  printf("\nMore details in:\n");
  printf("[1] T. Joachims, Making Large-Scale SVM Learning Practical. Advances in\n");
  printf("    Kernel Methods - Support Vector Learning, B. Schlkopf and C. Burges and\n");
  printf("    A. Smola (ed.), MIT Press, 1999.\n");
  printf("[2] T. Joachims, Estimating the Generalization performance of an SVM\n");
  printf("    Efficiently. International Conference on Machine Learning (ICML), 2000.\n");
  printf("[3] T. Joachims, Transductive Inference for Text Classification using Support\n");
  printf("    Vector Machines. International Conference on Machine Learning (ICML),\n");
  printf("    1999.\n");
  printf("[4] K. Morik, P. Brockhausen, and T. Joachims, Combining statistical learning\n");
  printf("    with a knowledge-based approach - A case study in intensive care  \n");
  printf("    monitoring. International Conference on Machine Learning (ICML), 1999.\n");
  printf("[5] T. Joachims, Learning to Classify Text Using Support Vector\n");
  printf("    Machines: Methods, Theory, and Algorithms. Dissertation, Kluwer,\n");
  printf("    2002.\n\n");
}


