/**
 * class for psvm
 */
package psvm;
import psvm.Result;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Collections;
import java.util.StringTokenizer;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.Math;

import libsvm.svm;
import libsvm.svm_model;
import libsvm.svm_node;
import libsvm.svm_parameter;
import libsvm.svm_print_interface;
import libsvm.svm_problem;
import psvm.ROC;

/**
 * This is the basic class for psvm
 */
public class SVM {
	DataSet ds_ = null;


	/**
	 * construct an empty svm class
	 */
	public SVM() { // create svm object
		ds_ = new DataSet();
	}


	/**
	 *  set default parameters
	 */
	public svm_parameter setDefaultParameter() {
		// default values
		svm_parameter param = new svm_parameter();

		param.svm_type = svm_parameter.C_SVC;
		param.kernel_type = svm_parameter.RBF;
		param.degree = 3;
		param.gamma = 0;	// 1/num_features
		param.coef0 = 0;
		param.nu = 0.5;
		param.cache_size = 100;
		param.C = 1;
		param.eps = 0.001;
		param.p = 0.1;
		param.shrinking = 0;
		param.probability = 0;
		param.nr_weight = 0;
		param.weight_label = new int[0];
		param.weight = new double[0];

		return param;
	}


	/** 
	 *  train model and save the model to file
	 */
	public svm_model trainModel(svm_problem prob, int b, double c, double j) {
		svm_parameter param = setDefaultParameter();

		param.svm_type = svm_parameter.C_SVC;	   // classification
		param.kernel_type = svm_parameter.RBF;  // linear kernel
		param.C = c;							   // cost 
		param.gamma = c;		
		param.probability = b;					   // out probability for prediction

		param.nr_weight = 2;					   // for two classes
		param.weight_label = new int[2];		   // two labels, 1 and -1
		param.weight_label[0] = 1;				   // set positive label
		param.weight_label[1] = -1;				   // set negative label

		param.weight = new double[2];			   // set weight for each class
		param.weight[0] = 1.0 / j;				   // set weight for positive class
		param.weight[1] = j;					   // set weight for negative class

		svm psvm = new svm();
		psvm.svm_check_parameter(prob, param);		   // check parameter before training, standard action
		svm_model model = psvm.svm_train(prob, param); // train the model

		return model;
	}


	/**
	 *  load model from file
	 */
	public svm_model loadModel(String modelFile) {
		svm psvm = new svm();
		svm_model sm = null;

		try { // read model from file
			sm = psvm.svm_load_model(modelFile); 

		} catch (IOException ex) {
			System.out.println(ex.getMessage());
			System.exit(1);
		}

		return sm;
	}


	/**
	 *  predict target value by given model and prob
	 */
	public double[] predictTarget(svm_model sm, svm_problem test) {
		svm psvm = new svm();
		double[] p = new double[test.l];

		for(int i = 0; i < test.l; i++) { // predict class label for each sample
			p[i] = svm.svm_predict(sm, test.x[i]);
		}

		return p;
	}


	/**
	 *  predict probability by given model and problem
	 */
	public double[] predictProbability(svm_model model, svm_problem prob) {
		svm psvm = new svm();
		int hasProb = svm.svm_check_probability_model(model);

		if(hasProb == 0) { // check the model has probability interface or not
			return null;
		}

		double[] p = new double[prob.l];
		double[] pp = new double[2]; // two class, need two probability estimation

		for(int i = 0; i < prob.l; i++) { // return the probability for each class
			
			// c is the class has the higher probability
			double c = psvm.svm_predict_probability(model, prob.x[i], pp);

			// get the probability of the positive class
			if (c == 1.0) {
				p[i] = Math.max(pp[0], pp[1]);  // get the higher probability
			} else {
				p[i] = Math.min(pp[0], pp[1]);  // get the lower probability
			}
		}

		return p;
	}


	/**
	 *  get target label from the probability output 
	 */
	public double[] getTargetByProbability(svm_model model, svm_problem prob) {
		double[] targets = predictProbability(model, prob);

		for(int i = 0; i < targets.length; i++) {
			double p = targets[i];
			if(p >= 0.5) {
				targets[i] = 1.0;
			} else {
				targets[i] = -1.0;
			}				
		}

		return targets;
	}


	/**
	 *  calculate accuracy of the model
	 */
	public class MyIntComparable implements Comparator<Result>{
		 
	    @Override
	    public int compare(Result a, Result b) {
	        return (a.prediction>b.prediction ? -1 : (a.prediction==b.prediction ? 0 : 1));
	    }
	}
	
	public double getAccuracy(double[] targets, svm_problem test) {
		List <Result> resultlist=new ArrayList <Result>();
		double true_positive=0;
		double true_negative=0;
		double false_positive=0;
		double false_negative=0;
		double total_positive=0;
		double total_negative=0;
		for(int i=0;i<test.l;i++){
			if(test.y[i]>0)
				total_positive++;
			else
				total_negative++;
			Result newnode=new Result(targets[i],test.y[i]);	
			resultlist.add(newnode);
		}
		Collections.sort(resultlist, new MyIntComparable());
		
		// ROC 
		
		List <ROC> roc=new ArrayList <ROC>();
		for (int i=0;i<test.l;i++){
			if(resultlist.get(i).original>0){
				true_positive++;
			}
			else{
				false_positive++;
			}
			true_negative=total_negative-false_positive;
			false_negative=total_positive-true_positive;
			roc.add(new ROC(true_positive,true_negative,false_positive,false_negative));
		}
		
		//AUC of the ROC curve.
		
		double sum=0;
		
		for (int i=1;i<test.l;i++){
			double TPR=roc.get(i).true_positive/total_positive;
			double FPR=roc.get(i).false_positive/total_negative;
			double TPR1=roc.get(i-1).true_positive/total_positive;
			double FPR1=roc.get(i-1).false_positive/total_negative;
			double TPR_ave=(TPR+TPR1)/2;
			double FPR_ave=(FPR-FPR1)/2;
			sum+=TPR_ave*FPR_ave;
		}

		
		double right=0;
		
		// count the TP in the top 20%
		
		for (int i=0;i<test.l*0.2;i++){
			if(resultlist.get(i).original>0){
				right+=1;
			}
		}
		// can return 20% TP, accuracy, or AUC of ROC.  Choose the best to specific problem.
		return right;
	}



	public double getRecall(double[] targets, svm_problem test) {
		double correct = 0.0;   // correct prediction number
		double plabel = 0.0;	// positive label number

		for(int i = 0; i < test.l; i++) {
			double label = test.y[i];
			double target = targets[i];

 			// calculate number of positive label
			if(label == 1.0) plabel += 1.0;

 			// calculate number of correct prediction
			if(label == target && label == 1.0) correct += 1.0;
		}

		// calculate recall
		return correct / plabel;
	}


	/**
	 *  calculate precision of the model
	 */
	public double getPrecision(double[] targets, svm_problem test) {
		double correct = 0.0;
		double plabel = 0.0;

		for(int i = 0; i < test.l; i++) {
			double label = test.y[i];
			double target = targets[i];

 			// calculate number of samples that classified as positive
			if(target == 1.0) plabel += 1.0;

 			// calculate number of correct prediction
			if(label == target && label == 1.0) correct += 1.0;
		}

		// calculate precision
		return correct / plabel;
	}


	/**
	 *  calculate F1 meausre of 
	 */
	public double getF1Measure(double[] targets, svm_problem test) {
		double precision = getPrecision(targets, test);
		double recall = getRecall(targets, test);
		return 2.0 * precision * recall / (precision + recall);
	}
	

	/**
	 *  train model with the given parameters and return the average accuracy of the model accuracy
	 */
	public double getModelAccuracy(svm_problem firstProb, svm_problem secondProb, double c, double j, int fold) {
		svm psvm = new svm();
		double accuracy = 0.0;

		// get the accuracy
		for(int i = 0; i < fold; i++) {
   			// train the first model
			svm_model sm = trainModel(firstProb, 0, c, j); // sv_coef 
			double[] targets = predictTarget(sm, secondProb);
			accuracy += getAccuracy(targets, secondProb);
		}

		return accuracy;
	}


	/**
	 *  get estimator c
	 */
	public double getC(svm_model sm, svm_problem prob) {
		double[] probC = predictProbability(sm, prob); // get the probability output by first model

		double c = 0.0;	  // estimator of p(s=1|y=1)
		double num = 0.0;

		for(int i = 0; i < prob.l; i++) {
			if(prob.y[i] == 1.0) { // samples, labeled and positive
				c += probC[i];
				num += 1.0;
			}
		}
		return c / num;
	}


	/**
	 *  weight sampls by their probability to class
	 */
	public svm_problem weightSamples(svm_model sm, svm_problem prob) {
		double[] probC = predictProbability(sm, prob); // get the probability output by first model
		
		for(int i=0;i<prob.l;i++){
			System.out.print(probC[i]+"\t");
			}
		
		double c = getC(sm, prob);
		Double C = (1.0 - c) / c;

		Vector<Integer>XX = new Vector<Integer>();	   // new smaples index
		Vector<Double>YY = new Vector<Double>();	   // new lables
		Vector<Double>weights = new Vector<Double>();  // weights for samples

		for(int i = 0; i < prob.l; i++) {
			if(prob.y[i] == 1.0) { // labeled and positive sample
				XX.add(i);		   // sample index
				YY.add(1.0);	   // set label to 1
				weights.add(1.0);  // set weight to 1

			} else {
				double Py1s0 = C * probC[i] / (1 - probC[i]);

				if(Py1s0 >= 1.0) { // if the sample's probability overflow
					XX.add(i);		   // sample index
					YY.add(1.0);	   // set label to 1
					weights.add(1.0);  // set weight to 1

				} else { // if the samples are not sure to be positive or negative
					XX.add(i);				  // sample index
					YY.add(1.0);			  // set label to positive
					weights.add(Py1s0);		  // set weight to p(y=1|s=0)
	
					XX.add(i);				  // sample index
					YY.add(-1.0);			  // set label to negative
					weights.add(1.0 - Py1s0); // set weight to 1 - p(y=1|s=0)
				}
			}
		}
		
		
		
		svm_problem newProb = new svm_problem(); 
		newProb.l = XX.size();	// set the samples size, the new size should be :    # of pos + 2* # of unlabeled

		// create new sample instance
		newProb.x = new svm_node[newProb.l][];
		newProb.y = new double[newProb.l];
		newProb.W = new double[newProb.l];

		for(int i = 0; i < newProb.l; i++) {
			newProb.x[i] = prob.x[XX.get(i)];  // samples
			newProb.y[i] = YY.get(i);		   // class label
			newProb.W[i] = weights.get(i);	   // weight for each sample
		}

		return newProb;
	}

	/**
	 *  calculate c parameter
	 */
	Vector<Double> getCCost() {
		double c =1e-6;
		Vector<Double>cost = new Vector<Double>();

 		// create 20 c parametes
		for(int i = -5; i < 5; i++) {
			c = c * Math.pow(2, i);
			cost.add(c);
		}

		return cost;
	}


	/**
	 *  calculate j parameter from training sample
	 */
	Vector<Double> getJCost(svm_problem prob) {
		double pNum = 0.0;
		double nNum = 0.0;

 		// get the count of positive and negative samples
		for(int i = 0; i < prob.l; i++) {
			if(prob.y[i] > 0.0) pNum += 1.0;  // number of positive samples
			if(prob.y[i] < 0.0) nNum += 1.0;  // number of negative samples
		}

		Vector<Double>cost = new Vector<Double>();

		// original ratio between positive and negative
		double ratio = pNum / nNum;

		// create 20 j cost parameters
		for(int i = -5; i <= 5; i++) {
			cost.add(ratio * Math.pow(2,i));
		}
		
		return cost;
	}


	/**
	 *  get optimal parameters
	 */
	Vector<Double> getOptimalParameter(svm_problem firstProb, svm_problem secondProb,
										Vector<Double>cCost, Vector<Double>jCost) {
	
		int cIndex = 0;
		int jIndex = 0;
		double maxAcc = 0.0;

		for(int i = 0; i < cCost.size(); i++) { // get the parameter that get best performance
			for(int j = 0; j < jCost.size(); j++) {
				// get accuracy by different c and j parameters, two folds validation
				double cc = getModelAccuracy(firstProb, secondProb, cCost.get(i), jCost.get(j), 2);

				if(maxAcc < cc) { // find c and j parameters that have the max accuracy 
					cIndex = i;
					jIndex = j;
					maxAcc = cc;
				}
			}
		}

		Vector<Double>cost = new Vector<Double>();
		cost.add(cCost.get(cIndex));  // c parameter
		cost.add(jCost.get(jIndex));  // j parameter

		return cost;
	}


	/**
	 *  output classification result
	 */
	public void saveResult(svm_problem testProb, double[] targets, String stage) {
		try {
			
			FileWriter out = new FileWriter(stage);
			// output classification result
			int correct = 0;
			for(int i = 0; i < testProb.l; i++) {
				if(testProb.y[i] == targets[i]) correct += 1;

				out.write(String.valueOf(testProb.y[i]));
				out.write(" : ");
				out.write(String.valueOf(targets[i]));
				out.write("\n");
			}

			out.close();

			/*
			System.out.println("\n");
			System.out.println(getAccuracy(targets, testProb));
			System.out.println(getPrecision(targets, testProb));
			System.out.println(getRecall(targets, testProb));
			System.out.println(getF1Measure(targets, testProb));
			*/
		} catch (IOException ex) {
			System.out.println(ex.getMessage());
			System.exit(1);
		}
	}


	/**
	 *  save to log file
	 * @throws UnsupportedEncodingException 
	 * @throws FileNotFoundException 
	 */
	public void saveLog(Vector<Double>param, String message) throws FileNotFoundException, UnsupportedEncodingException {
		PrintWriter writer = new PrintWriter("log", "UTF-8");
		writer.println("the c parameter is "+param.get(0));
		writer.println("the j parameter is "+param.get(1));
		writer.close();
	}




	/**
	 *  two stage trainning
	 */
	public void trainModel(String firstModel, String secondModel, String firstFile,
								 String secondFile) {
		try {
			svm psvm = new svm();

			// first stage model training: 
			svm_problem firstProb = ds_.readSamples(firstFile);   // read samples for training file

			// second stage model training:
			svm_problem secondProb = ds_.readSamples(secondFile); // read samples for dev file

			// read c tuning paramters from file
			Vector<Double>cCost = getCCost();

			// compute j tuning paramters   J is the value of positive/negative * k
			Vector<Double>jCost = getJCost(firstProb);

			// get optimal c and j parameter
			Vector<Double>cost = getOptimalParameter(firstProb, secondProb, cCost, jCost);

   			// train the first model
			svm_model sm = trainModel(firstProb, 1, cost.get(0), cost.get(1));

			// save first model
			psvm.svm_save_model(firstModel, sm);

	 		// weight the samples by the probability output of first model
			svm_problem prob = weightSamples(sm, secondProb);

			// save the samples and weight for testing, libsvm 3.1.2
	   		//ds_.saveProblem(prob, probFile, weightFile);

			saveLog(cost, "c and j parameter: ");

			jCost.clear();  // clear j parameter vector
			jCost.add(1.0); // add default j parameter
			cost.clear();   // clear c and j parameter vector

			// get optimal c and j parameter
			cost = getOptimalParameter(prob, firstProb, cCost, jCost);

			saveLog(cost, "c and j parameter: ");

			// train second model
			sm = trainModel(prob, 0, cost.get(0), cost.get(1));

			// save second model
			psvm.svm_save_model(secondModel, sm);

		} catch (IOException ex) {
			System.out.println(ex.getMessage());
			System.exit(1);
		}
	}


	/**
	 * test phase, loads model file and predict labels
	 */
	public void testSVM(String modelFile, String devFile, String testFile, String stage) {
		try {
			svm psvm = new svm();
			svm_problem testProb = ds_.readSamples(testFile);  // read test samples
			svm_model sm = psvm.svm_load_model(modelFile);	   // load svm model from file

			double[] targets = predictTarget(sm, testProb);	   // predict label of test samples

			saveResult(testProb, targets, stage);

		} catch (IOException ex) {
			System.out.println(ex.getMessage());
			System.exit(1);
		}
	}


	/**
	 *  main program
	 */
	static public void main(String[] args) {
		SVM svm = new SVM();
		if(args[0].equals("test")){
		//	svm.testSVM("first", "ABdev", "Ctest", "psvm.out");
			svm.testSVM(args[1],args[2],args[3],args[4]);
		}
		else if(args[0].equals("train")){
		//	svm.trainModel("first","second","ABtrain","ABdev");
			svm.trainModel(args[1],args[2],args[3],args[4]);
		}
		else if(args[0].equals("-h")){
			System.out.println("If you want to train the model, please input the following arguments: train $first_model_filename, $second_model_filename, $train_filename, $development_filename ");
			System.out.println("If you want to test with the mode, please input the following arguments: test $model_filename, $developement_filename, $test_filename, $output_filename");
		}
		

	}
}

